open Core.Std

module Input = struct
    type t = char list

    let of_string s = List.rev (String.to_list_rev s)
    let to_string = String.of_char_list
end

type 'result t = Input.t -> ('result * Input.t) option

let debug x name =
  fun input ->
  printf "Running %s on %s\n" name (Input.to_string input);
  let res = x input in
  begin match res with
	| Some (_, rest) -> printf "  Success; remaining: %s\n" (Input.to_string rest)
	| None -> printf "  Failed.\n"
  end;
  res

let return x = fun input -> Some (x, input)

let (>>=) x f = fun input ->
  match x input with
  | Some (y, input') -> f y input'
  | None -> None

let (<|>) x y = fun input ->
  match x input with
  | Some _ as res -> res
  | None -> y input

let any = function
  | [] -> None
  | (c::rest) -> Some (c, rest)

let fail _ = None

let satisfy p = any >>= fun c -> if p c then return c else fail

let one_of options = satisfy (fun c -> List.mem options c)

let eof = function
  | [] -> Some ((), [])
  | _ -> None

let (>>) x y = x >>= fun _ -> y

let between before after x = before >> x >>= fun r -> after >> return r

let rec many x = (x >>= fun r -> many x >>= fun rs -> return (r :: rs)) <|> return []
let many1 x = x >>= fun r -> many x >>= fun rs -> return (r :: rs)

let chain x op =
  let rec loop acc =
    (op >>= fun combine -> x >>= fun b -> loop (combine acc b)) <|> return acc
  in
  x >>= loop

let space         = satisfy Char.is_whitespace
let spaces        = many space >> return ()
let letter        = satisfy Char.is_alpha
let digit         = satisfy Char.is_digit
let alpha_num     = satisfy Char.is_alphanum
let non_alpha_num =
  satisfy (fun c -> not (Char.is_whitespace c
			 || Char.is_alphanum c
			 || c = '(' || c = ')'
			 || c = '{' || c = '}'))
let paren_char    = satisfy (fun c -> c = '(' || c = ')')
let brace_char    = satisfy (fun c -> c = '{' || c = '}')

let token_of char_p =
  spaces
  >> many1 char_p
  >>= fun cs ->
  spaces
  >> return (String.of_char_list cs)

let token_of_one char_p =
  spaces
  >> char_p
  >>= fun c ->
  spaces
  >> return (String.of_char c)

let token s =
  (token_of alpha_num <|> token_of non_alpha_num <|> token_of_one paren_char <|> token_of_one brace_char)
  >>= fun tok ->
  if tok = s then return () else fail

let parse p s =
  match p (Input.of_string s) with
  | Some (res, _) -> Some res
  | None -> None
