open Core.Std
open Miniparse

type expression =
  | Var of string
  | IntLit of int
  | BoolLit of bool
  | Neg of expression
  | Not of expression
  | Plus of expression * expression
  | Minus of expression * expression
  | Times of expression * expression
  | Divide of expression * expression
  | LT of expression * expression
  | GT of expression * expression
  | Equal of expression * expression
  | And of expression * expression
  | Or of expression * expression

type statement =
  | Assign of string * expression
  | If of expression * statement * statement
  | While of expression * statement
  | Block of statement list
  | Print of expression
  | Empty

type value = IntVal of int | BoolVal of bool

let reserved =
  [ "if"
  ; "then"
  ; "else"
  ; "while"
  ; "do"
  ; "and"
  ; "or"
  ; "not"
  ; "true"
  ; "false"
  ; "print"
  ]

let number = token_of digit >>= fun tok -> return (Int.of_string tok)
let identifier =
  token_of alpha_num
  >>= fun tok ->
  if String.length tok = 0 || not (Char.is_alpha tok.[0]) || List.mem reserved tok
  then fail
  else return tok

let parens = between (token "(") (token ")")
let braces = between (token "{") (token "}") 
let addop = token "+" >> return (fun x y -> Plus (x, y))
let subop = token "-" >> return (fun x y -> Minus (x, y))
let mulop = token "*" >> return (fun x y -> Times (x, y))
let divop = token "/" >> return (fun x y -> Divide (x, y))
let ltop  = token "<" >> return (fun x y -> LT (x, y))
let gtop  = token ">" >> return (fun x y -> GT (x, y))
let eqop  = token "==" >> return (fun x y -> Equal (x, y))
let orop  = token "or"  >> return (fun x y -> Or (x, y))
let andop = token "and" >> return (fun x y -> And (x, y))
let atom = (identifier >>= (fun s -> return (Var s)))
	   <|> (number >>= (fun x -> return (IntLit x)))
	   <|> (token "true" >> return (BoolLit true))
	   <|> (token "false" >> return (BoolLit false))

let rec expr input = (chain and_expr orop) input
and and_expr input = (chain not_expr andop) input
and not_expr input =
  let p = (token "not" >> rop_expr >>= fun e -> return (Not e)) <|> rop_expr in
  p input
and rop_expr input = (chain add_expr (ltop <|> gtop <|> eqop)) input
and add_expr input = (chain mul_expr (addop <|> subop)) input
and mul_expr input = (chain neg_expr (mulop <|> divop)) input
and neg_expr input =
  let p = (token "-" >> prm_expr >>= fun e -> return (Neg e)) <|> prm_expr in
  p input
and prm_expr input = (parens expr <|> atom) input

let rec stmt_seq input =
  let p =
    many1 atomic_stmt
    >>= fun ss ->
    match ss with
    | [s] -> return s
    | _ -> return (Block ss)
  in
  p input
and atomic_stmt input =
  let p =
    spaces
    >> (braces stmt_seq
	<|> if_stmt
	<|> while_stmt
	<|> print_stmt
	<|> assign_stmt)
  in
  p input
and if_stmt input =
  let p =
    token "if"
    >> expr
    >>= fun cond ->
    token "then"
    >> atomic_stmt
    >>= fun s1 ->
    (token "else" >> atomic_stmt) <|> return Empty
    >>= fun s2 ->
    return (If (cond, s1, s2))
  in
  p input
and while_stmt input =
  let p =
    token "while"
    >> expr
    >>= fun cond ->
    token "do"
    >> atomic_stmt
    >>= fun s ->
    return (While (cond, s))
  in
  p input
and print_stmt input =
  let p =
    token "print"
    >> expr
    >>= fun e ->
    token ";"
    >> return (Print e)
  in
  p input 
and assign_stmt input =
  let p =
    identifier
    >>= fun name ->
    token ":="
    >> expr
    >>= fun e ->
    token ";"
    >> return (Assign (name, e))
  in
  p input

let prog =
  stmt_seq >>= fun s ->
  spaces
  >> eof
  >> return s

let whole_expr =
  expr >>= fun s ->
  spaces
  >> eof
  >> return s

let parse_expr s = parse whole_expr s
let parse_string s = parse prog s
let parse_file file = parse prog (In_channel.read_all file)
